@@ -44,14 +44,8 @@ Write the parameters in ''p'' in a text format onto ''stream''.
 */
 void a_param_out_str(FILE *stream, a_param_ptr p);
 
-/*@manual aparam_internal
-TODO
-*/
 void a_param_inp_generic (a_param_ptr p, fetch_ops_t fops, void *ctx);
 
-/*@manual aparam_internal
-Initializes ''pairing'' with type A parameters in ''p''.
-*/
 void pairing_init_a_param(pairing_t pairing, a_param_t p);
 
 #endif //A_PARAM_H
